
#include  <SPI.h>
#include  <HighPowerStepperDriver.h>

HighPowerStepperDriver driver;


#define _DEBUG
#ifdef _DEBUG
#define Debugln(x) Serial.println(x)
#define Debug(x) Serial.print(x)
#define Here() Serial.println(__LINE__)
#else
#define Debugln(x)
#define Debug(x)
#define Here()
#endif // _DEBUG

#define _countof(x) (sizeof(x)/sizeof((x)[0]))
//#undef F
//#define F(x) x


#define STATUS(x, s) if(status & (1 << (uint8_t)(HPSDStatusBit::x))) \
                      {\
                      Serial.print(F("   0x"));\
                      Serial.print(1 << (uint8_t)HPSDStatusBit::x,HEX);\
                      Serial.println(F(" " #x " " s));\
                      }

static const int vrxAPin = A7;
static const int vryAPin = A6;
static const int swAPin =  7;
static const int vrxBPin = A4;
static const int vryBPin = A1;
static const int swBPin =  6;

static const int DriverSelectPin = 5;

static const int potPin = A0;

static  int fractionStep = 8;// step interval is 1/fractionStep
static  int stepsPerRevolution =200* fractionStep;
static const int moveFast = 500;
static const int moveSlow = 800;
static const int moveVariable = -1;
static char stepMode = ' ';
static int moveDirection = 1;
static  int minWait = 50;
static  int maxWait = 3000;
 
//static const int driftX = 8;
//static const int driftY = 8;
static bool useJoystick = true;
bool wantStatus = false;

static const int MAXTEST = 100;

/*************************************************************************
 * The nominal zero positions of the joysticks.  These will be recomputed
 * by the calibrate() procedure
 *************************************************************************/
static int vrxAZero = 508;
static int vryAZero = 516;

/**************************************************************************
 * class Joystick
 **************************************************************************/
class Joystick {
public:
  int XPin;  //analog input pin
  int YPin;  //analog input pin
  int swPin;  // switch pin

  int XZero;  // X Zero value 
  int YZero;  //Y Zero value

   int Xdrift;
   int Ydrift;

   Joystick ( int XP,int YP,int SPi,int XZ = 490,int YZ = 516,int XD = 8, int YD = 8)
   {
   XPin = XP;
   YPin = YP; 
   swPin = SPi;
   XZero = XZ;
   YZero = YZ;
   Xdrift = XD;
   Ydrift = YD;
   }

   void calibrate()
   {
    XZero = average(XPin);
    YZero = average(YPin);   
   }

   void showcalibrate()
   {
    Serial.print(F("X = "));
    Serial.print(analogRead(XPin));
    Serial.print(F(" XZero = "));
    Serial.print(XZero);
    Serial.print(F(" Xdrift = "));
    Serial.print(Xdrift);
    Serial.print(F(" ("));
    Serial.print(XZero - Xdrift);
    Serial.print(F(".."));
    Serial.print(XZero + Xdrift);
    Serial.print(F(") YZero = "));
    Serial.print(YZero);
   }

   /*******************************************************************
    * Joystick::getXWait
    * Inputs:
    *    bool & nothing: Set 'true' if there is nothing to do
    *    bool & direction: Set 'true' for positive motion, false for negative motion
    *    
    * Effect:
    * Computes the "wait time" to control the speed of the X-axis
    * This is done by reading the X-position potentiometer, and computing
    * the absolute value from the zero point
    *             X0
    * 0           |         1023
    *             V
    * -X0...................1023-X0
    *         |
    *********************************************************************/
   int getXWait(bool & nothing, bool &  direction)
   {
      int vrx = analogRead(XPin);
      int speed = map( vrx, 0, 1023, - XZero, 1023 - XZero);
      int abspeed = abs(speed);
      nothing = abspeed == 0;
      int limit = vrx < XZero ? XZero : 1023 - XZero;
      int wait = map (abspeed, 0, limit, maxWait, minWait);
      direction = speed > 0;
      return  wait;
   }

  
   int getX ()
    {
      return analogRead(XPin);
    }
   int getY ()
    {
      return analogRead(YPin);
    } 
   int getsw ()
    {
      return digitalRead(swPin);
    }
    /************************************************
     * inDriftX, inDriftY
     * Inputs:
     *      int x (or y): The x- (or y-) coordinate read from the joystick
     * Result: bool
     *       true if the value is within the nominal 
     *       "drift" of the zero  point
     ************************************************/
    bool inDriftX(int x )
    {
    return (abs(x - XZero) <= Xdrift); 
    }
        bool inDriftY ( int y )
    {
    return (abs(y - YZero) <= Ydrift);  
    }
protected:
    int average(int Pin)
    {
    long Center = 0;
    for(int i = 0; i < MAXTEST; ++ i)
        {
        Center += analogRead(Pin);  
        }
    return  Center / MAXTEST;     
    }
}; // class Joystick

Joystick JSleft(vrxAPin,vryAPin,swAPin);
Joystick JSright(vrxBPin,vryBPin,swBPin);

static bool debug = false;
static bool motordebug = false;
static bool showStep = false;
/***********************************************************
 * uSteps
 * 
 * This is the table of microstep values and the corresponding
 * command names.  This table will be printed out as part of the
 * 'help' message
 ***********************************************************/
typedef struct {
    char name;               // command letter
    int minWait;             // minimum delay in microseconds
    HPSDStepMode Mode;       // the value to use in setStepMode
    int fraction;            // human-readable fraction number
} MSTABLE;

MSTABLE uSteps17[] = {// Stepping table for nema 17
  //cmd  |minWait| Mode                      
    {'w',     530, HPSDStepMode::MicroStep1,     1},     //  [0]
    {'h',     220, HPSDStepMode::MicroStep2,     2},     //  [1]
    {'q',      70, HPSDStepMode::MicroStep4,     4},     //  [2]
    {'e',       0, HPSDStepMode::MicroStep8,     8},     //  [3]
    {'s',       0, HPSDStepMode::MicroStep16,   16},     //  [5]
    {'V',       0, HPSDStepMode::MicroStep32,   32},     //  [6]
    {'X',       0, HPSDStepMode::MicroStep64,   64},     //  [7]
    {'Y',       0, HPSDStepMode::MicroStep128, 128},     //  [8]
    {'Z',       0, HPSDStepMode::MicroStep256, 256},     //  [9]
    {0}         // EOT                                   // [10]
}; 

MSTABLE uSteps34[] = {// Stepping table for nema 34
  //cmd  |minWait| Mode                      | fraction
  //-----+-------|---------------------------+----------    
    {'w',    8000, HPSDStepMode::MicroStep1,     1},     //  [0]
    {'h',    6000, HPSDStepMode::MicroStep2,     2},     //  [1]
    {'q',    4000, HPSDStepMode::MicroStep4,     4},     //  [2]
    {'e',    2000, HPSDStepMode::MicroStep8,     8},     //  [3]
    {'s',       0, HPSDStepMode::MicroStep16,   16},     //  [5]
    {'V',       0, HPSDStepMode::MicroStep32,   32},     //  [6]
    {'X',       0, HPSDStepMode::MicroStep64,   64},     //  [7]
    {'Y',       0, HPSDStepMode::MicroStep128, 128},     //  [8]
    {'Z',       0, HPSDStepMode::MicroStep256, 256},     //  [9]
    {0}         // EOT                                   // [10]
}; 

MSTABLE * uSteps;

typedef struct {
    const char * name;
    MSTABLE * uSteps;
    int mA;
    // more fields TBD  
} MOTORS;

MOTORS motors[] = {
  {("NEMA-17"),  uSteps17,1500},
  {("NEMA-34"),  uSteps34,4200},
//{("NEMA-42"),  uSteps42,6000},
//{("NEMA-8"),   uSteps8,  600},
};

void showMotors()
{
  for(int i = 0; i < _countof(motors); ++i)
  {
    Serial.print(F("["));
    Serial.print(i);
    Serial.print(F("] ") );
    Serial.print(motors[i].name);
    Serial.print(F(" "));
    Serial.print(motors[i].mA);
    Serial.print(F("mA"));
    Serial.println();
    
      }
  
                                      
}


 static int offset = 0;

/*************************************************************
 * calibrate
 * 
 * Effect:
 *    Computes the "at rest" position of the joystick potentiometer
 *    It is nominally 512, but there is no guarantee
 * Notes:
 *    This is currently called by explicit user command.  In a
 *    user-input-free environment, this should probably be called
 *    from setup() to establish the current best-guess value
 **************************************************************/
 
 void calibrate()
 {
  Serial.println(F("CALIBRATION STARTING. DO NOT TOUCH JOYSTICK"));
  Serial.println(F("Current calibration"));
  JSleft.showcalibrate();
  Serial.println();

  JSleft.calibrate();
  //JSright.calibrate();
  

  Serial.print(F("Calibration complete"));
  Serial.println();
  JSleft.showcalibrate();
  //JSright.showcalibrate();
  Serial.println();
 } // calibrate

 /*******************************************************************
  * setStep
  * 
  * Inputs:
  *     char step: The step interval, one of the commands in the table
  * Result: bool
  *     true if the step mode was set successfully, or is already correct
  *     false if the step mode is not recognized
  ********************************************************************/
  
bool setStep(char step)
  {
   if(step == stepMode)
     return true; // not changed because already set

   // Look up the command in the table
   for(int i = 0; uSteps [i].name != 0 ; i++)
    { /* search step table */
      if(uSteps[i].name == step)
      { /* valid command */
        driver.setStepMode(uSteps[i]. Mode);
        stepMode = step;
        fractionStep = uSteps[i].fraction;
        stepsPerRevolution = 200 * fractionStep;
        minWait = uSteps [i].minWait + offset; 
        minWait = max(minWait,0);
        if(debug)
          { /* debug output */
            Serial.print(F("step size set to '"));
            Serial.print(step);
            Serial.print(F("' 1/"));
            Serial.print(fractionStep);
            Serial.println();
          } /* debug output */
        
        return true;  // successfully changed
       } /* valid command */
     } /* search step table */
  return false;
} // setStep

/****************************************************************************
 * getVariableWait
 * Inputs:
 *         bool & nothing: set to 'true' if nothing to be done
 *         bool & direction: 'true' for 'forward', 'false' for 'backward'
 * Result: int
 *         a variable Wait interval,in microseconds, based on the potentiometer
 *         a negative value indicates a reverse direction
 * Notes:
 *         In potentiometer mode, direction is set by the 'r' command and is ignored
 *         In potentiometer mode, nothing is meaningless and is not set
 *         
 *         'forward' and 'backward' are defined by how the motor is wired in
 *         (phase and phase orientation matter) and therefore cannot be said to
 *         be "clockwise" or "counterclockwise", but simply two directions which
 *         are opposite
 *****************************************************************************/
 
 int getVariableWait(bool & nothing, bool &  direction)
  {
   if (!useJoystick)  
     return map( analogRead(potPin), 0, 1023, minWait, maxWait);

   // For the joystick, read the joystick x,y values
   return JSleft.getXWait(nothing, direction);
 } // getVariableWait
 
/**********************************************************************
 * move
 * inputs:
 *   int motor: the motor number(1 = x, 2 = y,3 =z)
 *   long distance: number of steps to move ( + = cw - = ccw)
 *   int wait: The delay between steps
 *             If < 0, reads the potentiometer or joystick
 *             If > 0, the actual delay time in ms
 *   Notes:
 *       At the moment, since there is only one motor, the motor parameter
 *       is ignored.
 **********************************************************************/

void move(int motor, long distance, int Wait)
{
   driver.setDirection( distance >= 0 ?  true : false);
   
   if (motordebug)
   {
       if (abs(distance) < 2)
           {
           Serial.print(distance);
           }
   }
   // Serial.println(distance);
   for (long i = 0; i < abs(distance); i++) {
      if (Serial.available() )
         return;    // If  user has typed a command, abort the loop

      bool nothing;
      bool direction;
      
      int loopWait = Wait > 0 ? Wait : getVariableWait(nothing, direction);
      
      if( useJoystick)
        { /* joystick motion */
          static int lastloopWait = 0;
          static int lastnothing =-1;
          static int lastdirection = -1;
          
          if (debug && (nothing != lastnothing ||
                        direction != lastdirection ||
                        loopWait != lastloopWait))
              { /* trace */
                Serial.print(F("loopWait = "));
                Serial.print(loopWait);
                lastloopWait = loopWait;
                Serial.print(F(" nothing = "));
                Serial.print(nothing);
                lastnothing = nothing;
                Serial.print(F(" direction = "));
                Serial.print (direction);
                lastdirection = direction;
                
                Serial.println();
              } /* trace */
              
          if (nothing)
             return;
         driver.setDirection( direction ? true : false);
      }  /* joystick motion */

      if(motordebug)
      {
        static int count = 0;
        count ++;
        if(count % 25 == 0)
          Serial.println();
        Serial.print(direction ? F(">") : F("<"));
      }
       
      driver.step();
      uint8_t status = driver.readStatus();
      
      static uint8_t laststatus = 0;
      if(wantStatus && status != 0 && status != laststatus)
        {
        STATUS(STD,"stall detected");
        STATUS(STDLAT,"latched stall detected");
        driver.clearStatus();
        laststatus = status;
        }
      
      if (showStep) 
      {
      static int count = 0;  // Keep lines short
      count++;
      if(count % 50 == 0)    // if value modulo is zero, we have printed out enough on this line
        Serial.println(); 
      
        Serial.print(direction ? F(">") : F("<"));
      }
    }
} // move

/******************************************************************
 * setup
 * Initializes the various states required
 * 
 ******************************************************************/
void setup() {
  
  SPI.begin();  // Required for Pololu 36V4 driver board
  
  Serial.begin(115200);
  double T = ((double)millis()) / 1000.0;
  Serial.print(F("T = "));
  Serial.print(T, 3);
  Serial.println();
  
  pinMode(DriverSelectPin,OUTPUT);
  delay(1);
  //Serial.print(1);

  driver.setChipSelectPin(DriverSelectPin);
  //Serial.print(2);
  
  driver.resetSettings();
  driver.clearStatus();
  //Serial.print(3);
  
  driver.setCurrentMilliamps36v4(4000);
  //Serial.print(4);

  driver.enableDriver();

  pinMode(vrxAPin,INPUT);
  pinMode(vryAPin,INPUT);
  pinMode(swAPin,INPUT_PULLUP);

 Here();
  Debug("verifySettings()=");
  Debugln(driver.verifySettings());
  Here();

 uSteps =  uSteps34;
 

  Serial.println(F("type H for help"));
} // setup

static const int Wait = 500;
char ticket = 's';
bool changed = false;
bool query = false;
static const long shortRevs = 2;
static const long longRevs = 100;
static const int dt = 100;

/********************************************************************
 * help
 * 
 * Effect:
 *    Prints out the help message to remind us of the commands
 ********************************************************************/
void help()
{
  Serial.println(F(__FILE__)); 
  Serial.println(F(__DATE__ __TIME__));
  Serial.println(F("H  this message"));
  Serial.println("?  current status");
  Serial.print(F("+  add "));
  Serial.print(dt);
  Serial.println(F("us to step wait time"));
  Serial.print(F("-  subtract "));
  Serial.print(dt);
  Serial.println(F("us from step wait time"));
  Serial.println(F("c  calibrate joystick"));
  Serial.println(F("d  toggle debug Mode"));
  Serial.println(F("D  toggle show-step Mode"));
  Serial.println(F("F  clear faults"));
  Serial.println(F("m  toggle motor debug Mode"));
  Serial.println(F("M  motor selection"));
  if(! useJoystick)
    Serial.println ( F("j  set Joystick mode"));
  if(useJoystick)  
    Serial.println ( F("p  set potentiometer mode "));
  if(! useJoystick)
    Serial.println(F("r  reverse direction"));
  Serial.println(F("S  clear status"));
  Serial.print(F("t  set revolutions to "));
  Serial.println(shortRevs);
  Serial.print(F("T  set revolutions to "));
  Serial.println(longRevs);
  for(int i = 0; uSteps [i].name != 0 ; i++)
  {
    Serial.print(uSteps [i].name);
    Serial.print(F("  set 1/"));
    Serial.print(uSteps [i].fraction);
    Serial.println(F(" step")); 
  }
} // help

/********************************************************************
 * showStatus
 * Inputs:
 *    uint8_t status: The status bits from readStatus or readFaults
 * Effect:
 *    displays the bits as symbolic status
 ********************************************************************/

void showStatus(uint8_t status)
{
    if(status != 0)
    {
      /********************************************************************
       * STATUS
       * Inputs:
       *     x: The status flag shift value HPSDStatusBit::x
       *     s: The textual explanation of the bit
       * Effect:
       *     Generates the string 
       *          0xDD NAME explanation 
       *     for the bit in the status flag value 'status' at position x
       *     DD is the status bit in hexadecimal
       *     NAME is the name of the status bit, e.g., "OTS"
       *     explanation is the description of the status bit
       * Notes:
       *     The string provided as the s argument must not use F(); that
       *     is provided in the body of the macro
       *     The \ ***MUST*** be the last character on the line, nothing may
       *     follow it, even a space.
       *********************************************************************/
  
    STATUS(OTS,"overtemperature shutdown");
    STATUS(AOCP,"channel A overcurrent shutdown");
    STATUS(BOCP,"channel B overcurrent shutdown");
    STATUS(APDF,"channel A predriver fault"); 
    STATUS(BPDF,"channel B predriver fault");
    STATUS(UVLO,"undervoltage lockout");
    
    // Note that for faults, these two bits will always be zero
    STATUS(STD,"stall detected");
    STATUS(STDLAT,"latched stall detected");
         
    }
}

void showOffset()
{
  Serial.print(F( "Offset = "));
  Serial.print(offset);
  Serial.println();
  Serial.print(F("minWait = "));
  Serial.print(minWait);
  Serial.print(F(" offset + minWait = "));
  Serial.print( offset + minWait);
  Serial.println();
}

 void showQuery(bool query)
 {
   Serial.print(query ? F("status '") : F("changed '"));
    Serial.print(stepMode);
    Serial.print(F("' minWait = "));
    Serial.print(minWait);
    
    Serial.print(F(" vw="));
    bool nothing;
    bool direction;
    Serial.print(getVariableWait(nothing, direction));
    
    Serial.print(F(" direction = "));
    Serial.print(direction);
    
    Serial.print(useJoystick ? F(" Joystick mode") : F(" potentiometer mode"));
    
    uint8_t status = driver.readStatus();
    Serial.print(F(" status = 0x"));
    Serial.print(status,HEX);
    
    uint8_t faults = driver.readFaults();
    Serial.print(F(" faults = 0x"));
    Serial.print(faults,HEX);
    
    Serial.println();

    JSleft.showcalibrate();
    Serial.println();
    
    if(status != 0)
       { /* has status */
        Serial.println(F("Status/Faults"));
        showStatus(status);
       } /* has status */
 }
 static  long revs = 100;
/**************************************************************
 * loop
 * 
 * Effect:
 *    The main polling loop.  If there are characters at the serial
 *    port, process them first.
 ***************************************************************/
void loop() 
{
  if(Serial.available())
  {  /* serial pending */  
   char ch = Serial.read();
   if( ch <= ' ')
     return; // ignore spaces and control characters
     
   changed = true;
   query = false;
   
   switch(ch)
   { /* ch */
    case'?':
      query = true;
      break;
      
    case '+':
      offset += dt;
      showOffset();
      return;
      
   case '-':
      offset -= dt;
      if(offset < 0)
         offset = 0;
      showOffset();
      return;
      
   case 'c':
      calibrate();
      return; 
      
   case 'd':
      debug = !debug;
      Serial.print(F("debug "));
      Serial.println(debug ? F("ON"): F("OFF"));      
      return;

   case 'D':
      showStep = !showStep;
      Serial.print(F("show-step "));
      Serial.println(showStep ? F("ON"): F("OFF"));      
      return;
      
   case 'F':
        {
          driver. clearFaults();
  
          uint8_t faults = driver.readFaults();  // verify that they were cleared and not stuck
          Serial.println(F("Faults cleared"));
          showStatus(faults);
        }
      return;
      
   case 'j':
      useJoystick = true;
      return;
   case 'M':
      showMotors();
      return;  
    case 'm':
      motordebug = !motordebug;
      Serial.print(F("motordebug "));
      Serial.println(motordebug ? F("ON"): F("OFF"));      
      return;
   
   case 'p':
      useJoystick = false; 
      return;  
      
   case 'r':
      if(! useJoystick)
          {
           moveDirection = - moveDirection;
           changed = true;
          }
      return;
      
   case 'S':
      driver.clearStatus();
      Serial.println(F("status cleared"));
      {
        uint8_t status = driver.readStatus();
        showStatus(status);
      }
      return;  
      
   case't':
       revs = shortRevs;
       return;
       
   case'T':
       revs = longRevs;
       return;
       
   case 'H':
       help();
       return;

   case ' ':
       return;   
   default:
       if(ch < ' ')
         return;
      // It might be a step request, or complete nonsense
      ticket = ch;
      break;
   } /* ch */
  

  if(!query && !setStep(ticket))
    { /* complete nonsense */
      Serial.print(F("unknown command '"));
      Serial.print(ticket);
      Serial.print(F("'"));
      Serial.println();
      ticket = ' ';
      return;
    } /* complete nonsense */
  
    if(query || (debug && changed))
        { /* report status */
          showQuery(query);
         
        } /* report status */
  } /* serial pending */
  // Clear the changed and query flags
  
  changed = false;
  query = false;
  
  if (useJoystick)
  { /* joystick mode */
      static int lastvrxA = 0;
      static int lastvryA = 0;
      static int lastswA = -1;
      
      int vrxA = JSleft.getX();
      int vryA = JSleft.getY();
      int swA = JSleft.getsw();
      
      if(JSleft.inDriftX(vrxA))
      //  && JSleft.inDriftY(vryA))
         return;   // ignore values within the "drift range" of the joystick
     // Here();
      
      if (debug && (vrxA != lastvrxA ||
          vryA != lastvryA ||
          swA != lastswA))
          { /* debug trace */
          Serial.print(F("vrxA = "));
          Serial.print(vrxA);
          lastvrxA = vrxA;
                                                                                                                       
          Serial.print (F(" vryA = "));
          Serial.print ( vryA);
          lastvryA = vryA;
          
          Serial.print (F(" swA = "));
          Serial.print (swA);
          lastswA = swA;
          Serial.println(); 
          } /* debug trace */
          
      move(0, 1, moveVariable);
    } /* joystick mode */
  else
    { /* potentiomenter mode */
     move(0,moveDirection * revs * stepsPerRevolution, moveVariable);// will cause it to turn revs.
    } /* potentiomenter mode */
} // loop
